package greenfoot;

import android.graphics.BitmapShader;
import android.graphics.Paint;
import android.graphics.Shader.TileMode;
import java.lang.Class;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * World is the world that Actors live in. It is a two-dimensional grid of
 * cells.
 * 
 * <p>All Actor are associated with a World and can get access to the world object.
 * The size of cells can be specified at world creation time, and is constant
 * after creation. Simple scenarios may use large cells that entirely contain
 * the representations of objects in a single cell. More elaborate scenarios may
 * use smaller cells (down to single pixel size) to achieve fine-grained
 * placement and smoother animation.
 * 
 * <p>The world background can be decorated with drawings or images.
 * 
 * @see greenfoot.Actor
 * @author Upzealot
 * @version 0.1
 */

public class World
{
    final float Width;
	final float Height;
	final float cellSize;
	final float WidthPx;
	final float HeightPx;
	
	GreenfootImage background;
	final Paint bgPaint;
	
	
	List<Actor> actorList;
	List<Actor> actList;
	List<Actor> paintList;
	List<Class> classList4Act;
	List<Class> classList4Paint;
	//boolean sortA;
	//boolean sortP;
	
	//boolean isRunning;
	
	World(){
		Width = 40;
		Height = 60;
		cellSize = 10;
		WidthPx = Width * cellSize;
		HeightPx = Height * cellSize;

		actorList = new ArrayList<Actor>();
		actList = new ArrayList<Actor>();
		paintList = new ArrayList<Actor>();
		classList4Act = new ArrayList<Class>();
		//sortA = false;
		classList4Paint = new ArrayList<Class>();
		//sortP = false;
		background = new GreenfootImage((int)WidthPx, (int)HeightPx);
		background.setColor(255, 255, 255, 255);
		background.fill();

		bgPaint = new Paint();
		bgPaint.setShader(new BitmapShader(background.bmp, TileMode.REPEAT, TileMode.REPEAT));
	}
	
	/**
     * Construct a new world. The size of the world (in number of cells) and the
     * size of each cell (in pixels) must be specified.
     * 
     * @param w The width of the world (in cells).
     * @param h The height of the world (in cells).
     * @param cs Size of a cell in pixels.
     */
	public World(float w, float h, float cs){
		Width = w;
		Height = h;
		cellSize = cs;
		WidthPx = Width *cellSize;
		HeightPx = Height * cellSize;
		
		actorList = new ArrayList<Actor>();
		actList = new ArrayList<Actor>();
		paintList = new ArrayList<Actor>();
		classList4Act = new ArrayList<Class>();
		//sortA = false;
		classList4Paint = new ArrayList<Class>();
		//sortP = false;
		background = new GreenfootImage((int)WidthPx, (int)HeightPx);
		background.setColor(Color.WHITE);
		background.fill();
		background.setColor(Color.BLACK);
		
		bgPaint = new Paint();
		bgPaint.setShader(new BitmapShader(background.bmp, TileMode.REPEAT, TileMode.REPEAT));
	}
	
	void init(){
		//isRunning = false;
	}
	
	/**
     * Add an Actor to the world.
     * 
     * @param actor The new Actor to add.
     * @param x The x coordinate of the location where the object is added.
     * @param y The y coordinate of the location where the object is added.
     */
	private void addObject(Actor actor, float x, float y){
		actor.x = x;
		actor.y = y;
		actorList.add(actor);
		sortActOrder(actor);
		addToActList(actor);
		sortPaintOrder(actor);
		addToPaintList(actor);
		actor.world = this;
		actor.addedToWorld(this);
	}
	
	public void addObject(Actor actor, int X, int Y){
		addObject(actor, (float)X + 0.5f, (float)Y + 0.5f);
	}
	
	void addToActList(Actor actor){
		if(!actList.isEmpty())
		for(int i = actList.size() - 1; i >= 0; i--){
			if(actor.actOrder >= actList.get(i).actOrder){
				actList.add(i + 1, actor);
				return;
			}
		}
		actList.add(0, actor);
	}
	
	void addToPaintList(Actor actor){
		if(!paintList.isEmpty())
		for(int i = paintList.size() - 1; i >= 0; i--){
			if(actor.paintOrder >= paintList.get(i).paintOrder){
				paintList.add(i + 1, actor);
				return;
			}
		}
		paintList.add(0, actor);
	}
	
	/**
     * Remove an object from the world.
     * 
     * @param actor the object to remove
     */
	public void removeObject(Actor actor){
		actorList.remove(actor);
		actList.remove(actor);
		paintList.remove(actor);
		//actor.image.bmp.recycle();
	}
	
	/**
     * Remove a list of objects from the world.
     * 
     * @param actors A list of Actors to remove.
     */
    @SuppressWarnings("unchecked")
	public void removeObjects(Collection actors){
		Iterator it = actors.iterator();
		while(it.hasNext()){
			removeObject((Actor)it.next());
		}
	}
	
	/**
	 * Get the number of actors currently in the world.
	 * @return The number of actors
	 */
	public int numberOfObjects(){
		return actorList.size();
	}
	
	/**
     * Get all the objects in the world, or all the objects of a particular class.
     * <p>
     * If a class is specified as a parameter, only objects of that class (or
     * its subclasses) will be returned.
     * <p>
     * 
     * @param cls Class of objects to look for ('null' will find all objects).
     * 
     * @return A list of objects.
     */
    @SuppressWarnings("unchecked")
	public List getObjects(Class cls){
		if(cls == Actor.class || cls == null || actorList.isEmpty())
			return actorList;
		else{
			List<Object> aimList = new ArrayList<Object>();
			for(int i = 0; i < actorList.size(); i++){
				Actor actor = actorList.get(i);
				if(actor.isChildOf(cls))
					aimList.add((Object)actor);
			}
			return aimList;
		}
	}
	
	/**
     * Return all objects at a given cell.
     * <p>
     * 
     * An object is defined to be at that cell if its graphical representation
     * overlaps the center of the cell.
     * 
     * @param x X-coordinate of the cell to be checked.
     * @param y Y-coordinate of the cell to be checked.
     * @param cls Class of objects to look return ('null' will return all
     *            objects).
     */
    @SuppressWarnings("unchecked")
	public List getObjectsAt(int x, int y, Class cls){
		List<Object> aimList = new ArrayList<Object>();
		for(int i = 0; i < actorList.size(); i++){
		Actor actor = actorList.get(i);
		if(actor.isChildOf(cls)
			&& actor.getX() == x && actor.y == y)
			aimList.add((Object)actorList.get(i));
		}
		return aimList;
	}
	
	/**
	 * Return the size of a cell (in pixels).
	 */
	public int getCellSize(){
		return Math.round(cellSize);
	}
	float getCellSizeF(){
		return cellSize;
	}
	
	/**
	 * Return the width of the world (in number of cells).
	 */
	public int getWidth(){
		return Math.round(Width);
	}
	float getWidthF(){
		return Width;
	}
	
	/**
	 * Return the height of the world (in number of cells).
	 */
	public int getHeight(){
		return Math.round(Height);
	}
	float getHeightF(){
		return Height;
	}
	
	/**
     * Set a background image for the world. If the image size is larger than
     * the world in pixels, it is clipped. If it is smaller than the world, it
     * is tiled. A pattern showing the cells can easily be shown by setting a
     * background image with a size equal to the cell size.
     * 
     * @see #setBackground(String)
     * @param bg The image to be shown
     */
	public void setBackground(GreenfootImage bg){
		background = new GreenfootImage((int)WidthPx, (int)HeightPx);
		bgPaint.setShader(new BitmapShader(bg.bmp, TileMode.REPEAT, TileMode.REPEAT));
		background.canvas.drawRect(0, 0, WidthPx, HeightPx, bgPaint);
	}
	
	/**
     * Set a background image for the world from an image file. Images of type
     * 'jpeg', <s>'gif</s>' and 'png' are supported. If the image size is larger than
     * the world in pixels, it is clipped. A pattern showing the cells can
     * easily be shown by setting a background image with a size equal to the
     * cell size.
     * 
     * @see #setBackground(GreenfootImage)
     * @param filename The file holding the image to be shown
     * @throws IllegalArgumentException If the image can not be loaded.
     */
	public void setBackground(String filename){
		setBackground(new GreenfootImage(filename));
	}
	
	/**
     * Return the world's background image. The image may be used to draw onto
     * the world's background.
     * 
     * @return The background image
     */
	public GreenfootImage getBackground(){
		return background;
	}
	
	/**
     * Return the color at the centre of the cell. To paint a color, you need to
     * get the background image for the world and paint on that.
     *
     * @param x The x coordinate of the cell.
     * @param y The y coordinate of the cell.
     * @see #getBackground()
     * @throws IndexOutOfBoundsException If the location is not within the world
     *             bounds. If there is no background image at the location it
     *             will return Color.WHITE.
     */
	public Color getColorAt(int x, int y){
		return background.getColorAt((int)(x * cellSize + cellSize / 2), 
		                             (int)(y * cellSize + cellSize / 2));
	}
	
	/**
	 * Repaints the world.
	 */
	public void repaint(){
		Greenfoot.view.draw();
	}
	
	
	/**
     * Set the paint order of objects in the world. Paint order is specified
     * by class: objects of one class will always be painted on top of objects
     * of some other class. The order of objects of the same class cannot be 
     * specified.
     * Objects of classes listed first in the parameter list will 
     * appear on top of all objects of classes listed later.
     * <p>
     * Objects of a class not explicitly specified effectively inherit the paint
     * order from their superclass.
     * <p>
     * Objects of classes not listed will appear below the objects whose classes
     * have been specified.
     * 
     * @param classes  The classes in desired paint order
     */
    @SuppressWarnings("unchecked")
	public void setPaintOrder(Class... classes){
		classList4Paint = new ArrayList<Class>();
		for(int i = 0; i < classes.length; i++){
			classList4Paint.add(classes[i]);
		}
		sortPaintOrder();
	}
	
	void sortPaintOrder(){
		for(int i = 0; i < actorList.size(); i++)
		    sortPaintOrder(actorList.get(i));
		Collections.sort(paintList, new paintOrderComparator());
	}
	
	void sortPaintOrder(Actor actor){
		actor.paintOrder = 9999;
		Class curentClass = null;
		for(int i = 0; i < classList4Paint.size(); i++){
			Class class_i = classList4Paint.get(i);
			if(actor.isChildOf(class_i)
				&& (curentClass == null || curentClass.isAssignableFrom(class_i))){
				curentClass = class_i;
				actor.paintOrder = 9999 - i - 1;
			}
		}
	}
	
	
    /**
     * Set the act order of objects in the world. Act order is specified
     * by class: objects of one class will always act before objects
     * of some other class. The order of objects of the same class cannot be 
     * specified.
     * 
     * <p>Objects of classes listed first in the parameter list will 
     * act before any objects of classes listed later.
     * 
     * <p>Objects of a class not explicitly specified inherit the act
     * order from their superclass.
     * 
     * <p>Objects of classes not listed will act after all objects whose classes
     * have been specified.
     * 
     * @param classes
     *            The classes in desired act order
     */
    @SuppressWarnings("unchecked")
	public void setActOrder(Class... classes){
		classList4Act = new ArrayList<Class>();
		for(int i = 0; i < classes.length; i++){
			classList4Act.add(classes[i]);
		}
		sortActOrder();
	}
	
	void sortActOrder(){
		for(int i = 0; i < actorList.size(); i++)
		    sortActOrder(actorList.get(i));
		Collections.sort(actList, new actOrderComparator());
	}
	
	void sortActOrder(Actor actor){
		actor.actOrder = 9999;
		Class curentClass = null;
		for(int i = 0; i < classList4Act.size(); i++){
			Class class_i = classList4Act.get(i);
			if(actor.isChildOf(class_i)
				&& (curentClass == null || curentClass.isAssignableFrom(class_i))){
				curentClass = class_i;
				actor.actOrder = i;
			}
		}
	}
	
	
	/**
     * This method is called by the Greenfoot system when the execution has
     * started. This method can be overridden to implement custom behaviour when
     * the execution is started.
     * <p>
     * This default implementation is empty.
     */
	public void started(){
	}
	
	/**
     * This method is called by the Greenfoot system when the execution has
     * stopped. This method can be overridden to implement custom behaviour when
     * the execution is stopped.
     * <p>
     * This default implementation is empty.
     */
	public void stopped(){
	}
	
	/**
     * Act method for world. The act method is called by the greenfoot framework
     * at each action step in the environment. The world's act method is called
     * before the act method of any objects in the world.<p>
     * 
     * This method does nothing. It should be overridden in subclasses to
     * implement an world's action.
     */
    
	public void act(){
	}
	void act0(float dt){
	    act();
	}
}
